
// dep.h
// Copyright 2022 Matthew Rickard
// This file is part of dep

// dep is free software: you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation, either version 3 of the License, or
// (at your option) any later version.

// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.

// You should have received a copy of the GNU General Public License
// along with this program.  If not, see <https://www.gnu.org/licenses/>.

#ifndef DEP_H
#define DEP_H

#include <vector>
#include <string>
#include <map>
#include <set>
#include "AbstractProcess.h"
#include "CmdLine.h"
#include "ParseTree.h"

#define FILEFLAG_PSEUDOFILE            1
#define FILEFLAG_COMPARE               2
#define FILEFLAG_LOCAL_MODTIME_VALID   4
#define FILEFLAG_INCLUDE_MODTIME_VALID 8
#define FILEFLAG_UNCHANGED            16
#define FILEFLAG_FOUND_INCLUDES       32
#define FILEFLAG_STAT_CHANGED         64
#define FILEFLAG_CMD_CHANGED         128

#define RULEFLAG_MARKED           1
#define RULEFLAG_REBUILD          2
#define RULEFLAG_BUILDSTATE_VALID 4

// build states. The values come from the min() algorithm for
// calculating state from source state.

#define BUILDSTATE_SOURCES_STALE 0 // the file's sources are out of date; build them first
#define BUILDSTATE_STALE         1 // this file is out of date, but can be rebuilt
#define BUILDSTATE_BUILDING      2 // currently being built
#define BUILDSTATE_STAT_DIFF     3 // stat() differences indicate this file has changed
#define BUILDSTATE_REBUILT       4 // has been rebuilt
#define BUILDSTATE_OK            5 // stat() calls show this file is up to date
#define BUILDSTATE_REBUILT_SAME  6 // has been rebuilt, and has not changed
                                   // (BuildFile computes this from FILEFLAG_UNCHANGED)
#define BUILDSTATE_NO_SOURCES    7 // file exists, and has no source files

/*

The minimum build state of all sources tells you what you should do with a rule:

N Minimum source state  Up-to-date  Current state  Description
- --------------------  ----------  -------------  -------------------------------------
0 unknown               Don't care  unknown        you can't make a decision yet
1 stale                 Don't care  unknown        you should rebuild, but not yet
2 rebuilt-different     Don't care  stale          you should rebuild now
3 ok                    No          stale          you should rebuild now
3 ok                    Yes         ok             you should not rebuild
4 rebuilt-same          Don't care  ok             you should not rebuild
5 no sources            Don't care  ok             you should not rebuild

*/

struct EnvVar {
  std::string name, value;
};

struct Rule;
struct RuleSet;

struct IncludesData {
  std::string filename;
  //std::vector<std::string> folders;
  std::string workingDir;
};

struct BuildFile {
  BuildFile(RuleSet *ruleset);
  ~BuildFile();
  int getFileState();
  int calculateModTimes();
  void outputTo(std::ostream &os);
  int treeAddInclude(const std::string &source, RuleSet *ruleSet, int duplicateCheckIdx);
  void invalidateState();
  void markFile();
  void treeAddIncludes(const std::string &filename, const std::vector<std::string> *folders,
    std::fstream &is, const char *workingDir);

  std::string filename;
  std::vector<Rule *> readRuleVector;
  Rule *writeRule;
  int fileflag;
  int modTimeResult;
  struct timespec localModTime, includeModTime;
  long fileSize;
  std::vector<BuildFile *> includeVector;
  std::vector<BuildFile *> includedByVector;
  RuleSet *ruleset;
  IncludesData *includesData;
  std::vector<std::string> folders;
};

struct Rule {
  void addTarget(const std::string &target, const int flags, const char *sourceFile, int sourceLine);
  int addSource(const std::string &source);
  void addCommand(std::string command);
  int getRuleState(int debug = 0);
  int startx(int &resortFlag);
  void setFolder(const std::string &s);
  void markRule();
  Rule(RuleSet *ruleset);
  int nextCommandx(int &resortFlag);
  int markRebuiltx();
  void setState(int buildState);
  void invalidateState();
  void ruleAddIncludes(const std::string &filename, const std::vector<std::string> *folders,
    std::fstream &is, const std::string &dir);

  std::vector<BuildFile *> targetVector;
  std::vector<BuildFile *> sourceVector;
  std::vector<std::string> commandVector;
  std::string folder;
  RuleSet *ruleset;
  int ruleflag;
  int buildstate;
  Pipeline pipeline;
  std::map<std::string, std::string> backup; // This uses less memory than a string
                                             // in BuildFile, and is measurably faster
  int runCount;
  int ignoreExitCode;
};

std::string normalizePath(std::string a);

struct RuleSet {
  Rule *rulesetAddRule();
  int processCmd(int argc, const char *argv[], File *outputFile, const RuleSet *cacheset);
  int run(File *outputFile);
  RuleSet();
  int waitForRuleExitx(int &resortFlag);
  void outputTo(std::ostream &os);
  BuildFile *getBuildFile(const char *filename);
  void compareToCache(const RuleSet &cache);
  void combineWithCache(const RuleSet &cache);

  std::vector<Rule *> ruleVector;
  std::map<std::string, BuildFile *> fileMap;

  char optionTreeIncludes;
  char optionRuleIncludes;
  char optionDryrun;
  char optionRebuildMarked;
  char optionRebuildAll;

  short senseNewer;
  short senseModTime;
  short senseSize;
  short senseRule;
  short failed;

  int verbosity;
  int maxPipelines;
  int maxNonIncludes;
  std::map<std::string, Cmdspec *> syntaxMap;
  std::map<std::string, std::string> variableMap;

  File *outputFile;

  std::string findFile(const std::string &filename, const std::vector<std::string> *folders,
    std::fstream &is, const char *readingFilename, const char *workingDir) const;
  int confirmRuleChanges;
  std::ostringstream failedCommands;
};

void findLoop(std::vector<int> &v, std::vector<int> &result, int (*fn)(int a, int b));
void sortRuleVector(std::vector<Rule *> &ruleVector);
std::string joinFolderAndFile(const std::string &folder, const std::string &file);

#endif

